#!/usr/bin/env python
from __future__ import absolute_import, division, print_function

"""compare_directory.py: compare two directories, using regexes

XXX: want to keep track of all files in new directory
"""

## Copyright 2011-2013 Michael M. Hoffman <michael.hoffman@utoronto.ca>

import filecmp
from os import walk
from re import compile as re_compile, escape
import sys
import tarfile

from path import Path
from six import viewitems, viewkeys
from six.moves import zip

from segway._util import maybe_gzip_open, SEGWAY_ENCODING
from segway.version import __version__

# marker for matching anywhere in file
ANY_LINE_MARKER = b"(%__ANY_LINE__%)"
ANY_LINE_MARKER_LENGTH = len(ANY_LINE_MARKER)

def get_dir_filenames(dirname):
    if (not Path(dirname).exists()):
        raise IOError("Directory %s not found" % dirname)
    for dirbasename, dirnames, filenames in walk(dirname):
        dirbasepath = Path(dirbasename)
        relative_dirbasename = dirbasename.partition(dirname)[2]

        if relative_dirbasename.startswith("/"):
            relative_dirbasename = relative_dirbasename[1:]

        relative_dirpath = Path(relative_dirbasename)

        try:
            dirnames.remove(".svn")
        except ValueError:
            pass

        dirnames.sort()

        for filename in sorted(filenames):
            filename_relative = str(relative_dirpath / filename)
            # not really absolute, but more so than relative
            filename_absolute = str(dirbasepath / filename)
            yield filename_relative, filename_absolute

# regular expression unescape
re_unescape = re_compile(b"\(%.*?%\)")

def make_regex(text):
    """
    make regex, escaping things that aren't with (% %)
    """
    spans = [match.span() for match in re_unescape.finditer(text)]

    res = [b"^"]
    last_end = 0
    for start, end in spans:
        res.append(escape(text[last_end:start]))
        res.append(text[start + 2:end - 2])  # eliminate (% and %)
        last_end = end
    res.extend([escape(text[last_end:]), b"$"])
    
    return re_compile(b"".join(res))


def compare_file(template_filename, query_filename):
    # quick comparison without regexes
    if filecmp.cmp(template_filename, query_filename, shallow=False):
        return True  # files are identical, skip slow regex stuff

    # If the files are different find and report the differences based on
    # embedded regex criteria
    
    # Create a template held out set for lines that can match elsewhere 
    match_anywhere_regexs = set()

    res = True
    with maybe_gzip_open(template_filename, mode = "rb") as template_file:
        with maybe_gzip_open(query_filename, mode = "rb") as query_file:
            for line_number, lines in enumerate(
                    zip(template_file, query_file),
                    start=1):
                template_line = lines[0]
                # If the template line can be matched anywhere
                if template_line.startswith(ANY_LINE_MARKER):
                    # Create regex from marker-removed template line
                    # Save to held out match anywhere line set
                    match_anywhere_regexs.add(
                        make_regex(template_line[ANY_LINE_MARKER_LENGTH:])
                    )
                else:
                    re_template = make_regex(template_line)
                    match = re_template.match(lines[1])
                    if not match:
                        res = False
                        print("Line %d differences for %s" % (
                            line_number,
                            template_filename))
                        print("Different line:\n%s" % (lines[1]))

    # If there are lines that may be matched anywhere
    if match_anywhere_regexs:
        # Fo each line in the the query file
        with maybe_gzip_open(query_filename, mode = "rb") as query_file:
            for query_line in query_file:
                # If a query line matches the regex in the match anywhere set
                regex_matches = []
                for template_regex in match_anywhere_regexs:
                    match = template_regex.match(query_line)
                    if match:
                        regex_matches.append(template_regex)

                # Remove any matched regexs from the match anywhere set
                match_anywhere_regexs.difference_update(regex_matches)
                # If there are no more match anywhere regexs
                if not match_anywhere_regexs:
                    # Stop searching
                    break

    # If there any match anywhere lines that do not match the query file
    if match_anywhere_regexs:
        # Report failure
        res = False
        for regex in match_anywhere_regexs:
            print("Cannot find any match for {}\n".format(regex.pattern))

    return res


class TestCounter(object):
    def __init__(self):
        self.num_error = 0
        self.num_success = 0

    def error(self, msg):
        self.num_error += 1
        print(" %s" % msg, file=sys.stderr)

    def success(self):
        self.num_success += 1


def compare_directory(template_dirname, query_dirname):
    counter = TestCounter()
    query_filenames = dict(get_dir_filenames(query_dirname))

    different_files_list = []

    template_filenames = get_dir_filenames(template_dirname)
    for template_filename_relative, template_filename in template_filenames:
        re_template_filename_relative = make_regex(template_filename_relative.encode(SEGWAY_ENCODING))

        query_filenames_items = viewitems(query_filenames)
        for query_filename_relative, query_filename in query_filenames_items:
            if re_template_filename_relative.match(query_filename_relative.encode(SEGWAY_ENCODING)):
                del query_filenames[query_filename_relative]
                if compare_file(template_filename, query_filename):
                    counter.success()
                else:
                    counter.error("diff '%s' '%s'" % (template_filename,
                                                      query_filename))
                    different_files_list.append((template_filename,
                                                 query_filename))

                break
        else:
            counter.error("query directory missing %s" % template_filename)

    for query_filename_relative in viewkeys(query_filenames):
        counter.error("template directory missing %s"
                      % query_filename_relative)

    if counter.num_error == 0:
        msg = "PASS: %s and %s: %d files match" % (template_dirname,
                                                   query_dirname,
                                                   counter.num_success)
        print(msg, file=sys.stderr)
        return 0
    else:
        msg = "FAIL: %s and %s: %d files match;" \
              " %d files mismatch" % (template_dirname, query_dirname,
                                      counter.num_success, counter.num_error)

        base_directory_name = (Path(template_dirname)
                               .abspath().dirname().basename())
        tar_filename = base_directory_name + "-changes.tar.gz"

        with tarfile.open(tar_filename, "w:gz") as tar:
            print("Archiving in {}:".format(tar_filename))
            for template_filename, query_filename in different_files_list:
                tar.add(Path(template_filename).relpath())
                tar.add(Path(query_filename).relpath())
                print(template_filename, query_filename)

        print(msg, file=sys.stderr)
        return 1


def parse_options(args):
    from optparse import OptionParser

    usage = "%prog [OPTION]... TEMPLATEDIR QUERYDIR"
    version = "%%prog %s" % __version__
    parser = OptionParser(usage=usage, version=version)

    options, args = parser.parse_args(args)

    if not len(args) == 2:
        parser.error("incorrect number of arguments")

    return options, args


def main(args=sys.argv[1:]):
    options, args = parse_options(args)

    return compare_directory(*args)

if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
from __future__ import division

# Copyright 2009, 2011, 2012 Michael M. Hoffman <mmh1@washington.edu>

from math import ceil
import sys

from optbuild import OptionBuilder

from .._util import MB
from .common import _JobTemplateFactory

NATIVE_SPEC_PROG = OptionBuilder()  # do not run

# Reference: https://github.com/natefoo/slurm-drmaa
NATIVE_SPEC_DEFAULT = dict(
    nodes=1,  # Number of nodes to run on
    ntasks_per_node=1,  # Number of tasks per node
    cpus_per_task=1,  # Number of cpus per task
    mincpus=1,  # Min cpus per node
    requeue=True,  # Allow requeue on node failure or job preemption
    share=True  # Allow job allocation to be shared on nodes
)

CLEAN_SAFE_TIME = int(300 * 0.9)  # Slurm minimum job age is 300


class JobTemplateFactory(_JobTemplateFactory):
    def __init__(self, *args, **kwargs):
        _JobTemplateFactory.__init__(self, *args, **kwargs)

    def make_res_req(self, mem_usage, tmp_usage):
        # XXX: Unused tmp_usage for tmp storage requirements
        # Required minimum memory
        res_req = make_single_res_req("mem", mem_usage)

        return res_req

    def make_native_spec(self):
        return " ".join([self.native_spec,
                         make_native_spec(**NATIVE_SPEC_DEFAULT),
                         self.res_req])


def make_native_spec(*args, **kwargs):
    return " ".join(NATIVE_SPEC_PROG.build_args(args=args, options=kwargs))


def get_job_max_query_lifetime():
    return CLEAN_SAFE_TIME


def make_single_res_req(name, mem):
    # round up to the next mebibyte
    # http://www.clusterresources.com/torquedocs21/2.1jobsubmission.shtml#size

    # slurm native spec has form --arg=value, except when it's -a=value
    # see http://apps.man.poznan.pl/trac/slurm-drmaa
    return "--%s=%d" % (name, ceil(mem / MB))


def main(args=sys.argv[1:]):
    pass


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
from __future__ import absolute_import, division, print_function, with_statement

__version__ = "$Revision$"

# Copyright 2009, 2011-2014 Michael M. Hoffman <michael.hoffman@utoronto.ca>

from collections import defaultdict
from heapq import heappop, heappush
from os import environ, EX_TEMPFAIL
import sys
from time import sleep

from six import string_types

DRIVER_NAME_OVERRIDE = environ.get("SEGWAY_CLUSTER")

DRIVER_NAME_LOCAL = "local"

if DRIVER_NAME_OVERRIDE == DRIVER_NAME_LOCAL:
    from .local import ExitTimeoutException, JobState, Session
else:
    try:
        from drmaa import ExitTimeoutException, JobState, Session
    except (ImportError, RuntimeError):
        DRIVER_NAME_OVERRIDE = DRIVER_NAME_LOCAL  # no DRMAA available
        from .local import ExitTimeoutException, JobState, Session

from .._util import constant

MSG_JOB_ERROR = """
Submitted Job (%s) failed. Failed Job: %s.
For details, check error messages in %s.
See the Troubleshooting section of the Segway documentation."""

FAILED = JobState.FAILED
DONE = JobState.DONE

NA_FACTORY = constant("NA")

# min time to wait between checking job status
# XXX: should be an option
try:
    MIN_JOB_WAIT_SLEEP_TIME = float(environ["MIN_JOB_WAIT_SLEEP_TIME"])
except KeyError:
    MIN_JOB_WAIT_SLEEP_TIME = 3.0

## credited for min time so that there is some buffer when considering
## whether to submit more jobs or not

NOMINAL_MIN_JOB_WAIT_SLEEP_TIME = MIN_JOB_WAIT_SLEEP_TIME + 3
MAX_JOB_WAIT_SLEEP_TIME = 10  # max time to wait between checking job status
MAX_JOB_ATTEMPTS = 2

# these settings limit job queueing to 360 at once


def is_running_locally():
    """Returns True if not submitting to a cluster system"""
    return DRIVER_NAME_OVERRIDE == DRIVER_NAME_LOCAL


def get_driver_name(session):
    if DRIVER_NAME_OVERRIDE:
        return DRIVER_NAME_OVERRIDE

    drms_info = session.drmsInfo

    # XXX: find out what Son of Grid Engine and GridScheduler report
    if (drms_info.startswith("GE") or drms_info.startswith("SGE")
            or drms_info.startswith("UGE") or drms_info.startswith("OGS/GE")):
        return "sge" # XXX: should probably change to GE
    elif "Platform LSF" in drms_info:  # includes "IBM Platform LSF"
        return "lsf"
    elif drms_info.startswith("SLURM"):
        return "slurm"
    # not sure what PBS and PBS Pro return here.
    elif drms_info.startswith("Torque"):
        return "pbs"
    else:
        msg = ("unsupported distributed resource management system: %s"
               % drms_info)
        raise ValueError(msg)

# non-reentrant code
with Session() as _session:
    driver_name = get_driver_name(_session)

driver = __import__(driver_name, globals(), locals(), [driver_name], 1)
JobTemplateFactory = driver.JobTemplateFactory
make_native_spec = driver.make_native_spec
get_job_max_query_lifetime = driver.get_job_max_query_lifetime


class RestartableJob(object):
    def __init__(self, session, job_tmpl_factory, global_mem_usage,
                 mem_usage_key):
        self.session = session
        self.job_tmpl_factory = job_tmpl_factory

        # last trial index tried
        self.trial_index = -1
        # Count of the number of attempts for this job
        # Will always be submitted once before resubmission
        # Does not count for out of memory errors
        self.num_job_attempts = 1

        self.global_mem_usage = global_mem_usage
        self.mem_usage_key = mem_usage_key

        # (-num_segs, -num_frames) so minimum value will be largest
        # num_segs, num_frames
        self.sort_key = (-mem_usage_key[1], -mem_usage_key[2])

    def __repr__(self):
        return "<RestartableJob '%s'>" % self.job_tmpl_factory.template.jobName

    def __lt__(self, other):
        return self.sort_key < other.sort_key

    def run(self):
        job_tmpl_factory = self.job_tmpl_factory

        global_mem_usage = self.global_mem_usage
        mem_usage_key = self.mem_usage_key
        trial_index = global_mem_usage[mem_usage_key]

        #print >>sys.stderr, ("self.trial_index=%d; trial_index=%d"
        #                     % (self.trial_index, trial_index))

        # if this index was tried before and unsuccessful, increment
        # and set global_mem_usage, controlling for race conditions
        if self.trial_index == trial_index:
            with global_mem_usage.lock:
                choices = [global_mem_usage[mem_usage_key], trial_index + 1]
                trial_index = max(choices)

            global_mem_usage[mem_usage_key] = trial_index

        self.trial_index = trial_index

        job_template = job_tmpl_factory(trial_index)
        res = self.session.runJob(job_template)

        assert res

        res_req = job_tmpl_factory.res_req
        if not isinstance(res_req, string_types):
            res_req = " ".join(res_req)

        jobname = job_template.jobName

        # alert the user if they are running locally
        if is_running_locally():
            job_location = "running locally"
        else:
            job_location = "queued"

        print("%s %s: %s (%s)" % (job_location, res, jobname, res_req), file=sys.stderr)

        return res

    def free_job_template(self):
        # the JobTemplateFactory should delete its own job template object
        # when it's no longer needed
        self.job_tmpl_factory.delete_job_template(self.session)

class RestartableJobDict(dict):
    def __init__(self, session, job_log_file, *args, **kwargs):
        self.session = session
        self.job_log_file = job_log_file
        self.unqueued_jobs = []

        return dict.__init__(self, *args, **kwargs)

    def calc_sleep_time(self):
        # it's not a good idea to keep increasing the amount of sleep
        # time just because you have completed jobs. The problem is
        # that this can result in it taking more than
        # 'get_job_max_query_lifetime' seconds to check a job again

        # XXX: we should calculate this against the maximum number of
        # submitted jobs rather than the current number. But for now
        # we should just stick to MIN_JOB_WAIT_SLEEP_TIME

        # +1 is to avoid dividing by zero when len(self) is 0
        clean_safe_sleep_time = get_job_max_query_lifetime() / (len(self) + 1)

        return min(clean_safe_sleep_time, MAX_JOB_WAIT_SLEEP_TIME)

    def is_sleep_time_gt_min(self):
        sleep_time = self.calc_sleep_time()
        # maximum value: MAX_JOB_WAIT_SLEEP_TIME
        # minimum value: CLEAN_SAFE_TIME / (num jobs + 1)

        return sleep_time > NOMINAL_MIN_JOB_WAIT_SLEEP_TIME

    def _queue_unconditional(self, restartable_job):
        """queue unconditionally; don't do any checks

        if you unconditionally queue more jobs than you can poll on
        time, then we will lose track of jobs
        """
        jobid = restartable_job.run()
        self[jobid] = restartable_job

    def queue(self, restartable_job):
        # handle dry run case
        if restartable_job is None:
            return

        # protect against queuing more jobs than I want to poll
        if self.is_sleep_time_gt_min():
            self._queue_unconditional(restartable_job)
        else:
            heappush(self.unqueued_jobs, restartable_job)

    def queue_unqueued_jobs(self):
        while self.unqueued_jobs and self.is_sleep_time_gt_min():
            self._queue_unconditional(heappop(self.unqueued_jobs))

    def get_job_info_exit_status(self, job_info):
        if job_info.hasSignal:
            res = job_info.terminatedSignal
            # just in case this edge case ever happens
            if res == 0:
                return "hasSignal"

            return res
        elif job_info.wasAborted:
            return "wasAborted"
        elif job_info.hasExited:
            return job_info.exitStatus

        # this happens when the exit status is unknown
        return "noExit"

    def process_job(self, jobid, job_info):
        exit_status = self.get_job_info_exit_status(job_info)
        restartable_job = self[jobid]

        # Only resubmit job if out-of-memory is reported:
        # Check for EX_TEMPFAIL and also treat SIGKILL as out of memory

        # If the job queue is already full, this will probably
        # result in the job going to unqueued jobs for now
        if (exit_status == EX_TEMPFAIL or
           exit_status == "SIGKILL"):
            self.queue(restartable_job)
        # Else if the job had an error that wasn't due to memory
        elif exit_status != 0:
            # If the job has been submitted more than or equal to the max
            # amount of attempts allowed
            if restartable_job.num_job_attempts >= MAX_JOB_ATTEMPTS:
                # Raise a runtime error regarding the job
                job_template_factory = restartable_job.job_tmpl_factory
                job_name = job_template_factory.template.jobName
                error_filename = job_template_factory.error_filename
                # job will not be resubmitted, so free the job template
                restartable_job.free_job_template()
                raise RuntimeError(MSG_JOB_ERROR %
                                   (jobid, job_name, error_filename))
            # Otherwise
            else:
                # Increment the job attempt count
                restartable_job.num_job_attempts += 1
                # Resubmit
                self.queue(restartable_job)

        jobname = restartable_job.job_tmpl_factory.template.jobName

        prog, num_segs, num_frames = restartable_job.mem_usage_key

        resource_usage_orig = job_info.resourceUsage
        resource_usage = defaultdict(NA_FACTORY, resource_usage_orig)

        try:  # SGE
            maxvmem = resource_usage_orig["maxvmem"]
        except KeyError:  # non-SGE systems
            maxvmem = resource_usage["vmem"]

        cpu = resource_usage["cpu"]
        row = [jobid, jobname, prog, str(num_segs), str(num_frames),
               maxvmem, cpu, str(exit_status)]

        print(*row, sep="\t", file=self.job_log_file)
        self.job_log_file.flush()  # allow reading file now

        if exit_status == 0:
            # job will not be resubmitted, so free the job template
            restartable_job.free_job_template()
        del self[jobid]

    def wait(self):
        session = self.session
        jobids = list(self.keys())

        while jobids:
            # check each job individually
            for jobid in jobids:
                # XXX: should be an improved
                # RestartableJobDict.calc_sleep_time()
                sleep(MIN_JOB_WAIT_SLEEP_TIME)

                try:
                    job_info = session.wait(jobid, session.TIMEOUT_NO_WAIT)
                except ExitTimeoutException:
                    # job isn't done yet
                    continue

                self.process_job(jobid, job_info)
                self.queue_unqueued_jobs()

                # XXX: should be able to check
                # session.jobStatus(jobid) but this has problems
                # 1. it returns DONE even with non-zero exit status
                # is this an SGE or DRMAA bug? shouldn't a
                # non-zero exit status be a failure? or is that
                # just LSF?
                # 2. sometimes I can't get the jobStatus() of a completed job:
                # InvalidJobException: code 18: The job specified
                # by the 'jobid' does not exist. see versions prior to
                # SVN r425 for code

            jobids = list(self.keys())



#!/usr/bin/env python
from __future__ import division

"""include.py: save include file
"""

__version__ = "$Revision$"

## Copyright 2012 Michael M. Hoffman <michael.hoffman@utoronto.ca>

from ._util import Saver

class IncludeSaver(Saver):
    resource_name = "segway.inc.tmpl"
    copy_attrs = ["card_seg_countdown", "card_supervision_label", "max_frames",
                  "num_segs", "num_subsegs", "resolution", "ruler_scale", 
                  "supervision_label_range_size"]

    def make_mapping(self):
        num_segs = self.num_segs

        if isinstance(num_segs, slice):
            num_segs = "undefined\n#error must define CARD_SEG"

        resolution = self.resolution
        ruler_scale = self.ruler_scale
        if ruler_scale % resolution != 0:
            msg = ("resolution %d is not a divisor of ruler scale %d"
                   % (resolution, ruler_scale))
            raise ValueError(msg)
        ruler_scale_scaled = ruler_scale // resolution

        return dict(card_seg=num_segs,
                    card_subseg=self.num_subsegs,
                    card_presence=resolution+1,
                    card_segCountDown=self.card_seg_countdown,
                    card_supervisionLabel=self.card_supervision_label,
                    card_frameIndex=self.max_frames,
                    supervisionLabel_rangeSize=self.supervision_label_range_size,
                    ruler_scale=ruler_scale_scaled)




#!/usr/bin/env python
from __future__ import absolute_import, division, with_statement, print_function

"""
layer: convert flattened Viterbi BED files to a layered thick/thin bed file
"""

# Copyright 2009-2012 Michael M. Hoffman <michael.hoffman@utoronto.ca>

from collections import defaultdict
import re
import sys
from tempfile import NamedTemporaryFile

from numpy import array, diff, vstack
from optbuild import OptionBuilder_ShortOptWithEquals
from six import iterkeys, reraise, viewitems
from six.moves import map
from tabdelim import DictReader

from .bed import get_trackline_and_reader_native
from ._util import (BED_SCORE, BED_STRAND, get_label_color,
                    maybe_gzip_open, memoized_property, PassThroughDict,
                    SUFFIX_BED, SUFFIX_TAB)

from .version import __version__

BED_START = "0"
ACCEPTABLE_STRANDS = set(".+")

OFFSET_START = 0
OFFSET_END = 1
OFFSET_LABEL = 2

TRACKLINE_DEFAULT = ["track", 'description="segway-layer output"',
                     "visibility=full"]

BEDTOBIGBED_PROG = OptionBuilder_ShortOptWithEquals("bedToBigBed")

COLOR_DEFAULT = "128,128,128"

# XXX: don't need counter, can use length
class IncrementingDefaultDict(defaultdict):
    def __init__(self, *args, **kwargs):
        defaultdict.__init__(self, *args, **kwargs)
        self.counter = 0

    def __missing__(self, key):
        self[key] = value = self.counter
        self.counter += 1

        return value

class Tee(object):
    def __init__(self, *args):
        self._items = args
        self._exits = []

    ## code borrowed from Python 2.7 contextlib.nested
    def __enter__(self):
        items = self._items
        new_items = []
        exits = []

        for item in items:
            exit = item.__exit__
            enter = item.__enter__
            new_items.append(enter())
            exits.append(exit)

        self._items = new_items
        self._exits = exits

        return self

    def __exit__(self, *exc):
        while self._exits:
            exit = self._exits.pop()
            try:
                if exit(*exc):
                    exc = (None, None, None)
            except:
                exc = sys.exc_info()
        if exc != (None, None, None):
            # Don't rely on sys.exc_info() still containing
            # the right information. Another exception may
            # have been raised and caught by an exit method
            reraise(exc[0], exc[1], exc[2])

    def write(self, *args, **kwargs):
        for item in self._items:
            item.write(*args, **kwargs)

    def flush(self, *args, **kwargs):
        for item in self._items:
            item.flush(*args, **kwargs)

def make_comment_ignoring_dictreader(iterable, *args, **kwargs):
    return DictReader((item for item in iterable if not item.startswith("#")),
                      *args, **kwargs)

def load_mnemonics(filename):
    mnemonics = PassThroughDict()
    ordering = []

    if filename is not None:
        with open(filename) as infile:
            for row in make_comment_ignoring_dictreader(infile):
                mnemonics[row["old"]] = row["new"]
                ordering.append(row["old"])

    ordering.reverse()

    return mnemonics, ordering

def uniquify(seq):
    """
    http://stackoverflow.com/questions/480214/how-do-you-remove-duplicates-from-a-list-in-python-whilst-preserving-order

    http://www.peterbe.com/plog/uniqifiers-benchmark
    """
    seen = set()
    seen_add = seen.add
    return [x for x in seq if x not in seen and not seen_add(x)]

re_stem = re.compile(r"^(.+(?=\.)|[A-Za-z]+|)")
def get_stem(text):
    # returns empty string when there is no stem part
    return re_stem.match(text).group(0)

def recolor(mnemonics, labels):
    res = {}

    stem_colors = IncrementingDefaultDict()

    for label in labels:
        stem = get_stem(mnemonics[label])

        res[label] = get_label_color(stem_colors[stem])

    return res

def make_layer_filename(filename):
    """
    exported to run.py

    replace .bed from right side with .layered.bed
    """
    left, _, right = filename.rpartition(".bed")

    return ".layered.bed".join([left, right])

def make_csv(seq):
    return ",".join(map(str, seq))

def update_trackline(trackline, updates):
    if not trackline:
        trackline = TRACKLINE_DEFAULT[:]

    for word_index, word in enumerate(trackline):
        if word == "visibility=dense":
            trackline[word_index] = "visibility=full"

    for key, value in viewitems(updates):
        start = key + "="
        for word_index, word in enumerate(trackline):
            if word.startswith(start):
                trackline[word_index] = '%s="%s"' % (key, value)
                break

def get_color(datum, label, label_key):
    try:
        return datum.itemRgb
    except AttributeError:
        # assign color if it's missing
        try:
            label_int = int(label)
        except ValueError:
            label_int = label_key

        return get_label_color(label_int)

class Segmentation(defaultdict):
    """
    defaultdict of chromosomes:
    chromosome: list of runs
    run: list of segments
    segment: 3-tuple of ints
    """
    def __init__(self):
        defaultdict.__init__(self, list)
        self.colors = {}
        self.label_dict = IncrementingDefaultDict()

    def load(self, infilename):
        colors = self.colors
        label_dict = self.label_dict

        with maybe_gzip_open(infilename) as infile:
            self.trackline, reader = get_trackline_and_reader_native(infile)
            for datum in reader:
                try:
                    assert datum.strand in ACCEPTABLE_STRANDS
                except AttributeError:
                    pass # no strand

                label = datum.name
                label_key = label_dict[label]

                start = datum.chromStart
                chromosome = self[datum.chrom]

                try:
                    run = chromosome[-1]
                except IndexError:
                    run = []
                    chromosome.append(run)
                else:
                    if run[-1][OFFSET_END] != start:
                        run = []
                        chromosome.append(run)

                segment = (datum.chromStart, datum.chromEnd, label_key)
                run.append(segment)

                colors[label] = get_color(datum, label, label_key)

    def load_mnemonics(self, mnemonic_filename):
        self.mnemonics, self.ordering = load_mnemonics(mnemonic_filename)

    @memoized_property
    def labels_sorted(self):
        return uniquify(self.ordering + sorted(iterkeys(self.colors)))

    def update_trackline(self, updates):
        update_trackline(self.trackline, updates)

    def recolor(self):
        mnemonics = self.mnemonics

        # only do this if you have mnemonics
        if mnemonics:
            self.colors = recolor(mnemonics, self.labels_sorted)

    def save(self, outfilename, bigbed_outfilename):
        outfile = maybe_gzip_open(outfilename, "wt")

        if bigbed_outfilename:
            temp_file = NamedTemporaryFile(prefix=__package__, suffix=SUFFIX_BED, mode="wt")
            bigbed_infilename = temp_file.name
            outfile = Tee(outfile, temp_file)

        with outfile as outfile:
            ends = self.write(outfile)

            if bigbed_outfilename:
                outfile.flush()

                # XXX: refactor this into a function in _util
                # XXX: want to make a new wrapper to NamedTemporaryFile
                # that closes after one level of context, deletes after
                # the next
                # with MyNamedTemporaryFile() as temp_file:
                #     with temp_file
                #         print >>temp_file, "blah"
                #     print temp_file.name
                with NamedTemporaryFile(prefix=__package__, suffix=SUFFIX_TAB, mode="wt") as sizes_file:
                    for chrom, end in viewitems(ends):
                        print(chrom, str(end), sep="\t", file=sizes_file)

                    sizes_file.flush()
                    BEDTOBIGBED_PROG(bigbed_infilename, sizes_file.name, bigbed_outfilename)


    def write_trackline(self, outfile):
        # If self.trackline is None then TypeErrors occur when joining/printing.
        # This assertion can help single out trackline if it is the culprit
        assert(self.trackline)
        try:
            final_outfile = outfile._items[0]
        except AttributeError:
            final_outfile = outfile
        print(" ".join(self.trackline), file=final_outfile)

    def write(self, outfile):
        ends = {}
        mnemonics = self.mnemonics
        labels_sorted = self.labels_sorted
        label_dict = self.label_dict
        colors = self.colors

        self.write_trackline(outfile)

        for chrom, chromosome in sorted(viewitems(self)):
            for run in chromosome:
                segments = array(run)

                start = segments[0, OFFSET_START]
                end = segments[-1, OFFSET_END]
                ends[chrom] = end

                for label in labels_sorted:
                    label_key = label_dict[label]
                    color = colors.get(label, COLOR_DEFAULT)

                    # find all the rows for this label
                    segments_label_rows = segments[:, OFFSET_LABEL] == label_key

                    # extract just the starts and ends
                    segments_label = segments[segments_label_rows,
                                              OFFSET_START:OFFSET_END+1]

                    # pad end if necessary
                    segments_label_list = [segments_label]
                    if not len(segments_label) or segments_label[-1, OFFSET_END] != end:
                        # must be end-1 to end or UCSC gets cranky.
                        # unfortunately this results in all on at the
                        # right edge of each region
                        segments_label_list.append((end-1, end))

                    # pad beginning if necessary
                    if not len(segments_label) or segments_label[0, OFFSET_START] != start:
                        segments_label_list.insert(0, (start, start+1))

                    segments_label = vstack(segments_label_list)

                    # reverse offset by start
                    segments_label -= start

                    block_count = str(len(segments_label))

                    block_sizes = diff(segments_label).ravel()
                    block_sizes_str = make_csv(block_sizes)

                    block_starts = segments_label[:, 0]
                    block_starts_str = make_csv(block_starts)

                    # this just passes through the label itself if there
                    # are no mnemonics
                    mnemonic = mnemonics[str(label)]

                    row = [chrom, str(start), str(end), mnemonic, BED_SCORE,
                           BED_STRAND, str(start), str(end), color, block_count,
                           block_sizes_str, block_starts_str]

                    print(*row, sep="\t", file=outfile)

        return ends

def layer(infilename="-", outfilename="-", mnemonic_filename=None,
          trackline_updates={}, bigbed_outfilename=None, do_recolor=False):
    segmentation = Segmentation()
    segmentation.load_mnemonics(mnemonic_filename)
    segmentation.load(infilename)
    segmentation.update_trackline(trackline_updates)

    if do_recolor:
        segmentation.recolor()

    segmentation.save(outfilename, bigbed_outfilename)

def parse_options(args):
    from optplus import OptionParser

    usage = "%prog [OPTION]... [INFILE] [OUTFILE]"
    version = "%%prog %s" % __version__
    parser = OptionParser(usage=usage, version=version)
    parser.add_option("-b", "--bigBed", metavar="FILE",
                      help="specify bigBed output file")
    parser.add_option("-m", "--mnemonic-file", metavar="FILE",
                      help="specify tab-delimited input file with mnemonic "
                      "replacement identifiers for segment labels")
    parser.add_option("--no-recolor", action="store_false",
                      dest="recolor", default=True,
                      help="don't recolor labels")
    parser.add_option("-s", "--track-line-set", metavar="ATTR VALUE",
                      action="update", help="set ATTR to VALUE in track line",
                      default={})

    options, args = parser.parse_args(args)

    if not len(args) <= 2:
        parser.print_usage()
        sys.exit(1)

    return options, args

def main(args=sys.argv[1:]):
    options, args = parse_options(args)

    return layer(mnemonic_filename=options.mnemonic_file,
                 trackline_updates=options.track_line_set,
                 bigbed_outfilename=options.bigBed,
                 do_recolor=options.recolor,
                 *args)

if __name__ == "__main__":
    sys.exit(main())
